@@ -121,7 +121,7 @@ def integrate_adaptive(rhs, jac, y0, x0, xend, atol, rtol, dx0=.0,
             Whether to return error_weights, estimated_local_errors in info dict.
         'constraints': array
             Per component constraints 0.0: no constraint, 1.0: >=0, -1.0: <=0, 2.0: >0.0, -2.0: <0.0.
-        'max_steps_between_jac': int
+        'max_num_steps_between_jac': int
              Maximum  number  of  timesteps to wait before recomputation of the Jacobian or
              recommendation to update the preconditioner.
         'stab_lim_det': bool
@@ -251,7 +251,7 @@ def integrate_predefined(rhs, jac, y0, xout, atol, rtol, jac_type="dense",
             Whether to return error_weights, estimated_local_errors in info dict.
         'constraints': array
             Per component constraints 0.0: no constraint, 1.0: >=0, -1.0: <=0, 2.0: >0.0, -2.0: <0.0.
-        'max_steps_between_jac': int
+        'max_num_steps_between_jac': int
              Maximum  number  of  timesteps to wait before recomputation of the Jacobian or
              recommendation to update the preconditioner.
         'stab_lim_det': bool
