@@ -268,6 +268,9 @@ class Integrator{ // Thin wrapper class of CVode in CVODES
     long int mxsteps_;
 public:
     void *mem {nullptr};
+#if SUNDIALS_VERSION_MAJOR >= 6
+    std::shared_ptr<sundials::Context> ctx;
+#endif
     long int ny {0};
     int nq {0};
     int verbosity = 50;  // "50%" -- plenty of room for future tuning.
@@ -276,12 +279,23 @@ public:
     bool autonomous_exprs {false};
     bool record_order = false, record_fpe = false, record_steps = false, record_mxss = false;
     bool stab_lim_det_ {false}; // reinit at autorestart... (fixed in pycvodes2)
-    double time_rhs {0}, time_jac {0}, time_roots {0}, time_quads {0}, time_prec {0}, time_jtimes {0};
+    double time_rhs {0}, time_jac {0}, time_roots {0}, time_quads {0}, time_prec {0}, time_jtimes {0}, time_jtsetup {0};
     std::vector<int> orders_seen, fpes_seen;
     std::vector<double> steps_seen, mxss_seen;  // Conversion from float / long double not a problem.
-    Integrator(const LMM lmm, const IterType iter) {
+    Integrator(const LMM lmm, const IterType iter
+#if SUNDIALS_VERSION_MAJOR >= 6
+               , std::shared_ptr<sundials::Context> sun_ctx)
+        : ctx(sun_ctx)
+#else
+          )
+#endif
+ {
 #if SUNDIALS_VERSION_MAJOR >= 4
-        this->mem = CVodeCreate(static_cast<int>(lmm));
+     this->mem = CVodeCreate(static_cast<int>(lmm)
+# if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+# endif
+);
         this->iter_ = iter;
 #else
         this->mem = CVodeCreate(static_cast<int>(lmm), static_cast<int>(iter));
@@ -295,12 +309,15 @@ public:
         if (this->errfp)
             fclose(errfp);
 #if SUNDIALS_VERSION_MAJOR >= 3
-        if (this->y_)
+        if (this->y_) {
             N_VDestroy(this->y_);
-        if (this->A_)
+        }
+        if (this->A_) {
             SUNMatDestroy(this->A_);
-        if (this->LS_)
+        }
+        if (this->LS_) {
             SUNLinSolFree(this->LS_);
+        }
 #endif
 #if SUNDIALS_VERSION_MAJOR >= 4
         SUNNonlinSolFree(NLS_);
@@ -310,9 +327,14 @@ public:
     void init(CVRhsFn cb, realtype t0, N_Vector y) {
         this->ny = NV_LENGTH_S(y);
 #if SUNDIALS_VERSION_MAJOR >= 3
-        if (y_)
+        if (y_) {
             throw std::runtime_error("y_ already allocated");
-        y_ = N_VNew_Serial(NV_LENGTH_S(y));
+        }
+        y_ = N_VNew_Serial(NV_LENGTH_S(y)
+#if SUNDIALS_VERSION_MAJOR >= 6
+                                  , *ctx
+#endif
+);
         std::memcpy(NV_DATA_S(y_), NV_DATA_S(y), ny*sizeof(realtype));
 #else
         auto y_ = y;
@@ -329,10 +351,18 @@ public:
 #if SUNDIALS_VERSION_MAJOR >= 4
         int flag;
         if (this->iter_ == IterType::Newton) {
-            NLS_ = SUNNonlinSol_Newton(y_);
+            NLS_ = SUNNonlinSol_Newton(y_
+#if SUNDIALS_VERSION_MAJOR >= 6
+                                  , *ctx
+#endif
+);
         } else if (this->iter_ == IterType::Functional) {
             const int n_accel_vecs = 0; // number of Anderson acceleration vectors (TODO: customizable)
-            NLS_ = SUNNonlinSol_FixedPoint(y_, n_accel_vecs);
+            NLS_ = SUNNonlinSol_FixedPoint(y_, n_accel_vecs
+#if SUNDIALS_VERSION_MAJOR >= 6
+                                  , *ctx
+#endif
+);
             flag = CVodeSetNonlinearSolver(this->mem, NLS_);
             if (flag == CV_SUCCESS) {
                 ; // pass
@@ -350,7 +380,11 @@ public:
         this->init(cb, t0, y.n_vec);
     }
     void init(CVRhsFn cb, realtype t0, const realtype * const y, int ny) {
-        SVector yvec (ny, const_cast<realtype*>(y));
+        SVector yvec (ny, const_cast<realtype*>(y)
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+);
         this->init(cb, t0, yvec.n_vec);
         // it is ok that yvec is destructed here
         // (see CVodeInit in cvodes.c which at L848 calls cvAllocVectors (L3790))
@@ -369,7 +403,11 @@ public:
         reinit(t0, y.n_vec);
     }
     void reinit(realtype t0, const realtype * const y, int ny){
-        SVector yvec (ny, const_cast<realtype*>(y));
+        SVector yvec (ny, const_cast<realtype*>(y)
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+);
         reinit(t0, yvec.n_vec);
     }
     // Root finding
@@ -394,7 +432,11 @@ public:
             check_flag(flag);
     }
     void quad_init(CVQuadRhsFn fQ, std::vector<realtype>& yQ0){
-        SVector yQ0_(yQ0.size(), yQ0.data());
+        SVector yQ0_(yQ0.size(), yQ0.data()
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+);
         quad_init(fQ, yQ0_.n_vec);
     }
     void quad_reinit(const N_Vector yQ0){
@@ -410,7 +452,11 @@ public:
         quad_reinit(yQ0.n_vec);
     };
     void quad_reinit(std::vector<realtype>& yQ0){
-        SVector yQ0_(yQ0.size(), yQ0.data());
+        SVector yQ0_(yQ0.size(), yQ0.data()
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+);
         quad_reinit(yQ0_.n_vec);
     }
     // Main solver optional input functions
@@ -471,7 +517,11 @@ public:
             throw std::runtime_error("CVodeSVtolerances failed.");
     }
     void set_tol(realtype rtol, std::vector<realtype> &atol){
-        SVector atol_(atol.size(), atol.data());
+        SVector atol_(atol.size(), atol.data()
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+);
         set_tol(rtol, atol_.n_vec);
     }
     void set_quad_err_con(bool errconQ){ // quad_init must have been called
@@ -508,7 +558,11 @@ public:
     //     set_quad_tol(reltolQ, abstolQ.n_vec);
     // }
     void set_quad_tol(realtype reltolQ, std::vector<realtype> &abstolQ){
-        SVector atol_(abstolQ.size(), abstolQ.data());
+        SVector atol_(abstolQ.size(), abstolQ.data()
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+);
         set_quad_tol(reltolQ, atol_.n_vec);
     }
     // set stop time
@@ -542,7 +596,11 @@ public:
         if (A_ == nullptr){
             if (A_)
                 throw std::runtime_error("matrix already set");
-            A_ = SUNDenseMatrix(ny, ny);
+            A_ = SUNDenseMatrix(ny, ny
+# if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+# endif
+);
             if (!A_)
                 throw std::runtime_error("SUNDenseMatrix failed.");
         }
@@ -556,16 +614,37 @@ public:
 #    else
                 SUNDenseLinearSolver
 #    endif
-                (y_, A_);
+                (y_, A_
+#   if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#   endif
+);
 #  else
-            LS_ = SUNLapackDense(y_, A_);
+            LS_ =
+#    if SUNDIALS_VERSION_MAJOR >= 4
+                SUNLinSol_LapackDense
+#    else
+                SUNLapackDense
+#    endif
+                (y_, A_
+#   if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#   endif
+);
 #  endif
             if (!LS_)
                 throw std::runtime_error("SUNDenseLinearSolver failed.");
         }
-        status = CVDlsSetLinearSolver(this->mem, LS_, A_);
-        if (status < 0)
+        status =
+#if SUNDIALS_VERSION_MAJOR >= 6
+            CVodeSetLinearSolver
+#else
+            CVDlsSetLinearSolver
+#endif
+            (this->mem, LS_, A_);
+        if (status < 0) {
             throw std::runtime_error("CVDlsSetLinearSolver failed.");
+        }
 #else
 #  if PYCVODES_NO_LAPACK == 1
         status = CVDense(this->mem, ny);
@@ -594,13 +673,27 @@ public:
         ){
         int status;
 #if SUNDIALS_VERSION_MAJOR >= 3
-        status = CVDlsSetJacFn(this->mem, djac);
-        if (status < 0)
-            throw std::runtime_error("CVDlsSetJacFn failed.");
+        status =
+#if SUNDIALS_VERSION_MAJOR >= 6
+            CVodeSetJacFn
 #else
-        status = CVDlsSetDenseJacFn(this->mem, djac);
-        if (status < 0)
+            CVDlsSetJacFn
+#endif
+(this->mem, djac);
+        if (status < 0) {
+            throw std::runtime_error("CVDlsSetJacFn failed.");
+        }
+#else
+        status =
+#if SUNDIALS_VERSION_MAJOR >= 6
+            CVodeSetDenseJacFn
+#else
+            CVDlsSetDenseJacFn
+#endif
+            (this->mem, djac);
+        if (status < 0) {
             throw std::runtime_error("CVDlsSetDenseJacFn failed.");
+        }
 #endif
     }
 
@@ -616,6 +709,9 @@ public:
 #  if SUNDIALS_VERSION_MAJOR < 4
                                , mlower+mupper
 #  endif
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
                 );
             if (!A_)
                 throw std::runtime_error("SUNDenseMatrix failed.");
@@ -630,7 +726,11 @@ public:
                 # else
                 SUNBandLinearSolver
                 #endif
-                (y_, A_);
+                (y_, A_
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+);
 #  else
             LS_ =
                 # if SUNDIALS_VERSION_MAJOR >= 4
@@ -638,14 +738,25 @@ public:
                 #else
                 SUNLapackBand
                 #endif
-                (y_, A_);
+                (y_, A_
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+);
 #  endif
             if (!LS_)
                 throw std::runtime_error("SUNDenseLinearSolver failed.");
         }
-        status = CVDlsSetLinearSolver(this->mem, LS_, A_);
-        if (status < 0)
+        status =
+#if SUNDIALS_VERSION_MAJOR >= 6
+            CVodeSetLinearSolver
+#else
+            CVDlsSetLinearSolver
+#endif
+            (this->mem, LS_, A_);
+        if (status < 0) {
             throw std::runtime_error("CVDlsSetLinearSolver failed.");
+        }
 #else
         status =
 #  if PYCVODES_NO_LAPACK == 1
@@ -666,7 +777,9 @@ public:
     }
 
     void set_band_jac_fn(
-#if SUNDIALS_VERSION_MAJOR >= 3
+#if SUNDIALS_VERSION_MAJOR >= 6
+        CVLsJacFn
+#elif SUNDIALS_VERSION_MAJOR >= 3
         CVDlsJacFn
 #else
         CVDlsBandJacFn
@@ -676,13 +789,21 @@ public:
 #if SUNDIALS_VERSION_MAJOR >= 3
         if (A_ == nullptr || LS_ == nullptr)
             throw std::runtime_error("set_linear_solver_to_banded not called?");
-        status = CVDlsSetJacFn(this->mem, djac);
-        if (status < 0)
+        status =
+# if SUNDIALS_VERSION_MAJOR >= 6
+            CVodeSetJacFn
+# else
+            CVDlsSetJacFn
+# endif
+            (this->mem, djac);
+        if (status < 0) {
             throw std::runtime_error("CVDlsSetJacFn failed.");
+        }
 #else
         status = CVDlsSetBandJacFn(this->mem, djac);
-        if (status < 0)
+        if (status < 0) {
             throw std::runtime_error("CVDlsSetBandJacFn failed.");
+        }
 #endif
     }
 
@@ -695,22 +816,46 @@ public:
     int status;
 #if SUNDIALS_VERSION_MAJOR >= 3
         if (A_ == nullptr){
-            if (A_)
-                throw std::runtime_error("matrix already set");
-            A_ = SUNSparseMatrix(ny, ny, nnz, CSC_MAT);
+            if (A_) {
+              throw std::runtime_error("matrix already set");
+            }
+            A_ = SUNSparseMatrix(ny, ny, nnz, CSC_MAT
+#   if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#   endif
+);
             if (!A_)
                 throw std::runtime_error("SUNSparseMatrix failed.");
         }
         if (LS_ == nullptr){
-            if (LS_)
+            if (LS_) {
                 throw std::runtime_error("linear solver already set");
-            LS_ = SUNKLU(y_, A_);
-            if (!LS_)
+            }
+            LS_ =
+#   if SUNDIALS_VERSION_MAJOR >= 4
+              SUNLinSol_KLU
+#   else
+              SUNKLU
+#   endif
+                (y_, A_
+#   if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#   endif
+              );
+            if (!LS_) {
                 throw std::runtime_error("SUNKLU failed.");
+            }
         }
-        status = CVDlsSetLinearSolver(this->mem, LS_, A_);
-        if (status < 0)
+        status =
+#   if SUNDIALS_VERSION_MAJOR >= 6
+            CVodeSetLinearSolver
+#   else
+            CVDlsSetLinearSolver
+#   endif
+            (this->mem, LS_, A_);
+        if (status < 0) {
             throw std::runtime_error("CVDlsSetLinearSolver failed.");
+        }
 #else
         status = CVKLU(this->mem, ny, nnz, CSC_MAT);
         if (status != CVSLS_SUCCESS) {
@@ -729,9 +874,16 @@ public:
         djac) {
         int status;
 #if SUNDIALS_VERSION_MAJOR >= 3
-        status = CVDlsSetJacFn(this->mem, djac);
-        if (status < 0)
+        status =
+#if SUNDIALS_VERSION_MAJOR >= 6
+            CVodeSetJacFn
+#else
+            CVDlsSetJacFn
+#endif
+            (this->mem, djac);
+        if (status < 0) {
             throw std::runtime_error("CVDlsSetJacFn failed.");
+        }
 #else
         status = CVSlsSetSparseJacFn(this->mem, djac);
         if (status < 0)
@@ -749,21 +901,45 @@ public:
         switch (solver) {
         case IterLinSolEnum::GMRES:
 #if SUNDIALS_VERSION_MAJOR >= 3
+#if SUNDIALS_VERSION_MAJOR >= 4
+            LS_ = SUNLinSol_SPGMR(y_, static_cast<int>(ptyp), maxl
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+);
+#else
             LS_ = SUNSPGMR(y_, static_cast<int>(ptyp), maxl);
+#endif
 #else
             flag = CVSpgmr(this->mem, static_cast<int>(ptyp), maxl);
 #endif
             break;
         case IterLinSolEnum::BICGSTAB:
 #if SUNDIALS_VERSION_MAJOR >= 3
+#if SUNDIALS_VERSION_MAJOR >= 4
+            LS_ = SUNLinSol_SPBCGS(y_, static_cast<int>(ptyp), maxl
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+);
+#else
             LS_ = SUNSPBCGS(y_, static_cast<int>(ptyp), maxl);
+#endif
 #else
             flag = CVSpbcg(this->mem, static_cast<int>(ptyp), maxl);
 #endif
             break;
         case IterLinSolEnum::TFQMR:
 #if SUNDIALS_VERSION_MAJOR >= 3
+#if SUNDIALS_VERSION_MAJOR >= 4
+            LS_ = SUNLinSol_SPTFQMR(y_, static_cast<int>(ptyp), maxl
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+);
+#else
             LS_ = SUNSPTFQMR(y_, static_cast<int>(ptyp), maxl);
+#endif
 #else
             flag = CVSptfqmr(this->mem, static_cast<int>(ptyp), maxl);
 #endif
@@ -812,7 +988,13 @@ public:
     }
     void set_jtimes_fn(CVSpilsJacTimesVecFn jtimes){
 #if SUNDIALS_VERSION_MAJOR >= 3
-        int flag = CVSpilsSetJacTimes(this->mem, NULL, jtimes);
+        int flag =
+# if SUNDIALS_VERSION_MAJOR >= 6
+            CVodeSetJacTimes
+# else
+            CVSpilsSetJacTimes
+# endif
+            (this->mem, NULL, jtimes);
 #else
         int flag = CVSpilsSetJacTimesVecFn(this->mem, jtimes);
 #endif
@@ -830,11 +1012,23 @@ public:
     }
 #endif
     void set_preconditioner(CVSpilsPrecSetupFn setup_fn, CVSpilsPrecSolveFn solve_fn){
-        int flag = CVSpilsSetPreconditioner(this->mem, setup_fn, solve_fn);
+        int flag =
+#if SUNDIALS_VERSION_MAJOR >= 6
+            CVodeSetPreconditioner
+#else
+            CVSpilsSetPreconditioner
+#endif
+            (this->mem, setup_fn, solve_fn);
         this->cvspils_check_flag(flag);
     }
     void set_iter_eps_lin(realtype delta){
-        int flag = CVSpilsSetEpsLin(this->mem, delta);
+        int flag =
+#if SUNDIALS_VERSION_MAJOR >= 6
+            CVodeSetEpsLin
+#else
+            CVSpilsSetEpsLin
+#endif
+            (this->mem, delta);
         this->cvspils_check_flag(flag, true);
     }
     void set_prec_type(PrecType pretyp){
@@ -842,11 +1036,23 @@ public:
 #if SUNDIALS_VERSION_MAJOR >= 3
         switch(solver_){
         case IterLinSolEnum::GMRES:
+#if SUNDIALS_VERSION_MAJOR >= 4
+            flag = SUNLinSol_SPGMRSetPrecType(LS_, (int)pretyp); break;
+#else
             flag = SUNSPGMRSetPrecType(LS_, (int)pretyp); break;
+#endif
         case IterLinSolEnum::BICGSTAB:
+#if SUNDIALS_VERSION_MAJOR >= 4
+            flag = SUNLinSol_SPBCGSSetPrecType(LS_, (int)pretyp); break;
+#else
             flag = SUNSPBCGSSetPrecType(LS_, (int)pretyp); break;
+#endif
         case IterLinSolEnum::TFQMR:
+#if SUNDIALS_VERSION_MAJOR >= 4
+            flag = SUNLinSol_SPTFQMRSetPrecType(LS_, (int)pretyp); break;
+#else
             flag = SUNSPTFQMRSetPrecType(LS_, (int)pretyp); break;
+#endif
         default:
             throw std::runtime_error("unknown solver kind.");
         }
@@ -859,7 +1065,11 @@ public:
         int flag;
 #if SUNDIALS_VERSION_MAJOR >= 3
         if (solver_ == IterLinSolEnum::GMRES)
+#if SUNDIALS_VERSION_MAJOR >= 4
+            flag = SUNLinSol_SPGMRSetGSType(LS_, (int)gs_type);
+#else
             flag = SUNSPGMRSetGSType(LS_, (int)gs_type);
+#endif
         else
             throw std::runtime_error("Setting Gram-Schmidt type only makes sense for GMRES");
 #else
@@ -874,9 +1084,19 @@ public:
         case IterLinSolEnum::GMRES:
             throw std::runtime_error("GMRES has no max length option");
         case IterLinSolEnum::BICGSTAB:
-            flag = SUNSPBCGSSetMaxl(LS_, maxl); break;
+#if SUNDIALS_VERSION_MAJOR >= 4
+            flag = SUNLinSol_SPBCGSSetMaxl(LS_, maxl);
+#else
+            flag = SUNSPBCGSSetMaxl(LS_, maxl);
+#endif
+            break;
         case IterLinSolEnum::TFQMR:
-            flag = SUNSPTFQMRSetMaxl(LS_, maxl); break;
+#if SUNDIALS_VERSION_MAJOR >= 4
+            flag = SUNLinSol_SPTFQMRSetMaxl(LS_, maxl);
+#else
+            flag = SUNSPTFQMRSetMaxl(LS_, maxl);
+#endif
+            break;
         default:
             throw std::runtime_error("unknown solver kind.");
         }
@@ -886,13 +1106,18 @@ public:
         this->cvspils_check_flag(flag);
     }
 // Linear solver options:
-    void set_max_steps_between_jac(long int msbj) {
+    void set_max_num_steps_between_jac(long int msbj) {
+#if SUNDIALS_VERSION_MAJOR >= 5 && SUNDIALS_VERSION_MINOR >= 4
+        int flag = CVodeSetJacEvalFrequency(this->mem, msbj+1);
+        cvspils_check_flag(flag);
+#else
 #if SUNDIALS_VERSION_MAJOR >= 4
         int flag = CVodeSetMaxStepsBetweenJac(this->mem, msbj);
         cvspils_check_flag(flag);
 #else
         ignore(msbj);
-        throw std::runtime_error("set_max_steps_between_jac  requires sundials >=4.0.0");
+        throw std::runtime_error("set_max_num_steps_between_jac  requires sundials >=4.0.0");
+#endif
 #endif
     }
 // Getters:
@@ -918,7 +1143,11 @@ public:
         this->get_err_weights(ew.n_vec);
     }
     void get_err_weights(realtype * ew) const {
-        SVectorV ew_(ny, ew);
+        SVectorV ew_(ny, ew
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+);
         this->get_err_weights(ew_.n_vec);
     }
     void get_est_local_errors(N_Vector ele) const {
@@ -928,7 +1157,11 @@ public:
         this->get_est_local_errors(ele.n_vec);
     }
     void get_est_local_errors(realtype * ele) const {
-        SVectorV ele_(ny, ele);
+        SVectorV ele_(ny, ele
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+);
         this->get_est_local_errors(ele_.n_vec);
     }
     void set_constraints(N_Vector constraints) {
@@ -950,7 +1183,11 @@ public:
 #endif
     }
     void set_constraints(const std::vector<realtype> &constraints) {
-        SVector sv_constraints(constraints.size(), constraints.data());
+        SVector sv_constraints(constraints.size(), constraints.data()
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+);
         set_constraints(sv_constraints.n_vec);
     }
 
@@ -958,7 +1195,13 @@ public:
     long int get_n_lin_iters() const {
         long int res=0;
         int flag;
-        flag = CVSpilsGetNumLinIters(this->mem, &res);
+        flag =
+#if SUNDIALS_VERSION_MAJOR >= 6
+            CVodeGetNumLinIters
+#else
+            CVSpilsGetNumLinIters
+#endif
+            (this->mem, &res);
         this->cvspils_check_flag(flag);
         return res;
     }
@@ -966,7 +1209,13 @@ public:
     long int get_n_prec_evals() const {
         long int res=0;
         int flag;
-        flag = CVSpilsGetNumPrecEvals(this->mem, &res);
+        flag =
+#if SUNDIALS_VERSION_MAJOR >= 6
+            CVodeGetNumPrecEvals
+#else
+            CVSpilsGetNumPrecEvals
+#endif
+            (this->mem, &res);
         this->cvspils_check_flag(flag);
         return res;
     }
@@ -974,7 +1223,13 @@ public:
     long int get_n_prec_solves() const {
         long int res=0;
         int flag;
-        flag = CVSpilsGetNumPrecSolves(this->mem, &res);
+        flag =
+#if SUNDIALS_VERSION_MAJOR >= 6
+            CVodeGetNumPrecSolves
+#else
+            CVSpilsGetNumPrecSolves
+#endif
+(this->mem, &res);
         this->cvspils_check_flag(flag);
         return res;
     }
@@ -982,7 +1237,13 @@ public:
     long int get_n_conv_fails() const {
         long int res=0;
         int flag;
-        flag = CVSpilsGetNumConvFails(this->mem, &res);
+        flag =
+#if SUNDIALS_VERSION_MAJOR >= 6
+            CVodeGetNumLinConvFails
+#else
+            CVSpilsGetNumConvFails
+#endif
+            (this->mem, &res);
         this->cvspils_check_flag(flag);
         return res;
     }
@@ -990,7 +1251,13 @@ public:
     long int get_n_jac_times_evals() const {
         long int res=0;
         int flag;
-        flag = CVSpilsGetNumJtimesEvals(this->mem, &res);
+        flag =
+#if SUNDIALS_VERSION_MAJOR >= 6
+            CVodeGetNumLinIters
+#else
+            CVodeGetNumJtimesEvals
+#endif
+            (this->mem, &res);
         this->cvspils_check_flag(flag);
         return res;
     }
@@ -998,7 +1265,13 @@ public:
     long int get_n_iter_rhs() const {
         long int res=0;
         int flag;
-        flag = CVSpilsGetNumRhsEvals(this->mem, &res);
+        flag =
+#if SUNDIALS_VERSION_MAJOR >= 6
+            CVodeGetNumRhsEvals
+#else
+            CVSpilsGetNumRhsEvals
+#endif
+            (this->mem, &res);
         this->cvspils_check_flag(flag);
         return res;
     }
@@ -1078,14 +1351,26 @@ public:
     }
     long int get_n_dls_jac_evals() const {
         long int res=0;
-        int flag = CVDlsGetNumJacEvals(this->mem, &res);
+        int flag =
+#if SUNDIALS_VERSION_MAJOR >= 6
+            CVodeGetNumJacEvals
+#else
+            CVDlsGetNumJacEvals
+#endif
+            (this->mem, &res);
         cvdls_check_flag(flag);
         return res;
     }
 
     long int get_n_dls_rhs_evals() const {
         long int res=0;
-        int flag = CVDlsGetNumRhsEvals(this->mem, &res);
+        int flag =
+#if SUNDIALS_VERSION_MAJOR >= 6
+            CVodeGetNumRhsEvals
+#else
+            CVDlsGetNumRhsEvals
+#endif
+(this->mem, &res);
         cvdls_check_flag(flag);
         return res;
     }
@@ -1129,10 +1414,11 @@ public:
             check_flag(flag);
         }
     }
-    void call_rhs(realtype t, SVector y, SVector &ydot){
+    void call_rhs(realtype t, SVector& y, SVector &ydot){
         int status = this->cb_(t, y.n_vec, ydot.n_vec, this->user_data_);
-        if (status)
+        if (status) {
             throw std::runtime_error("call_rhs failed.");
+        }
     }
     void unsuccessful_step_throw_(int flag){
         throw std::runtime_error(StreamFmt() << std::scientific << "Unsuccessful step (t="
@@ -1158,9 +1444,21 @@ public:
                  realtype ** ew_ele=nullptr // length(ew_ele) must be == 2*td*ny if not nullptr
         ){
         int status;
-        SVector y {ny};
-        SVector yQ {nq};
-        SVector work {ny};
+        SVector y {ny
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+};
+        SVector yQ {nq
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+};
+        SVector work {ny
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+};
         long int mxsteps = get_max_num_steps();
         if (*td < 1)
             throw std::logic_error("Expected td >= 1");
@@ -1193,7 +1491,11 @@ public:
             this->set_stab_lim_det(stab_lim_det_);
         }
         if (nq){
-            auto yQ0 = SVector(nq, &qout(tidx, 0));
+            auto yQ0 = SVector(nq, &qout(tidx, 0)
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+);
             this->quad_reinit(yQ0);
         }
         if (nderiv >= 1){
@@ -1257,23 +1559,29 @@ public:
                             std::clog << "cvodes_cxx.hpp:" << __LINE__ << ":" << __FUNCTION__ << " Autorestart (" << autorestart << ") t=" << cur_t
                                       << " nstp= " << get_n_steps() << " nfev=" << get_n_rhs_evals()  << " ae=" << ((autonomous_exprs) ? 't' : 'f') << "\n";
                             if (status >= 0) {
-                                N_Vector ele, ew;
-                                ele = N_VNew_Serial(ny);
-                                ew = N_VNew_Serial(ny);
-                                get_est_local_errors(ele);
-                                get_err_weights(ew);
-                                realtype mx = 0.0;
-                                int mxi = -1;
-                                for (unsigned i=0; i < ny; ++i){
-                                    const realtype cur = NV_DATA_S(ele)[i]*NV_DATA_S(ew)[i];
-                                    if (cur > mx){
-                                        mxi = i;
-                                        mx = cur;
-                                    }
-                                }
-                                N_VDestroy_Serial(ele);
-                                N_VDestroy_Serial(ew);
-                                std::clog << "cvodes_cxx.hpp:" << __LINE__ << ":" << __FUNCTION__ << ":     max(ew[i]*ele[i]) = " << mx << ", i=" << mxi << "\n";
+                              SVector ele {
+                                  ny
+#if SUNDIALS_VERSION_MAJOR >= 6
+                                  , *ctx
+#endif
+                              }, ew {
+                                  ny
+#if SUNDIALS_VERSION_MAJOR >= 6
+                                  , *ctx
+#endif
+                              };
+                              get_est_local_errors(ele);
+                              get_err_weights(ew);
+                              realtype mx = 0.0;
+                              int mxi = -1;
+                              for (unsigned i = 0; i < ny; ++i) {
+                                  const realtype cur = ele[i] * ew[i];
+                                  if (cur > mx) {
+                                      mxi = i;
+                                      mx = cur;
+                                  }
+                              }
+                              std::clog << "cvodes_cxx.hpp:" << __LINE__ << ":" << __FUNCTION__ << ":     max(ew[i]*ele[i]) = " << mx << ", i=" << mxi << "\n";
                             }
                         }
                         if (status == CV_CONV_FAILURE && autorestart == 1 && this->autorestart_relax_tolerances_last_restart) {
@@ -1282,7 +1590,7 @@ public:
                             this->set_tol(1e-3, 1e-3); if (this->verbosity > 0) std::clog << " - using atol=1e-3, rtol=1e-3";
                             this->set_dense_jac_fn(nullptr); if (this->verbosity > 0) std::clog << " - using finite differences.\n"; // Hail Mary
                         }
-                        if (this->verbosity > 0) std::cerr << '\n';
+                        if (this->verbosity > 0) std::clog << '\n';
                         if (tidx > mxsteps){
                             this->set_max_num_steps(mxsteps + autorestart_additional_steps);
                         }
@@ -1314,7 +1622,7 @@ public:
             for (int i=0; i<ny; ++i) {
 #if PYCVODES_CLIP_TO_CONSTRAINTS == 1
                 if (constraints_.size() and constraints_[i] == 1.0 and y[i] < 0) {
-                    if (this->verbosity > 60) { std::clog << "clipping y[" << i << "] to zero.\n"; }
+                    if (this->verbosity > 40) { std::clog << "clipping y[" << i << "] to zero.\n"; }
                     yout(tidx, 0, i) = 0.0;
                 } else {
                     yout(tidx, 0, i) = y[i];
@@ -1357,6 +1665,9 @@ public:
                 *xyqout = (realtype *)new_xyqout;
             }
         }
+        if (this->verbosity > 0) {
+            std::fflush(stderr);
+        }
         return tidx;
     }
 #undef datalen
@@ -1378,9 +1689,21 @@ public:
         int iout = 0;
         realtype cur_t=tout[0];
         int status;
-        SVector y {ny};
-        SVector yQ {nq};
-        SVector work {ny};
+        SVector y {ny
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+};
+        SVector yQ {nq
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+};
+        SVector work {ny
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+};
         long int mxsteps = get_max_num_steps();
         for (int i=0; i<ny; ++i)
             y[i] = yq0[i];
@@ -1389,7 +1712,11 @@ public:
             this->set_stab_lim_det(stab_lim_det_);
         }
         if (nq){
-            auto yQ0 = SVector(nq, const_cast<realtype*>(yq0) + (nderiv+1)*ny);
+            auto yQ0 = SVector(nq, const_cast<realtype*>(yq0) + (nderiv+1)*ny
+#if SUNDIALS_VERSION_MAJOR >= 6
+                   , *ctx
+#endif
+);
             this->quad_reinit(yQ0);
         }
         if (record_order)
@@ -1469,22 +1796,27 @@ public:
                         std::clog << "cvodes_cxx.hpp:" << __LINE__ << ":" << __FUNCTION__ << ": Autorestart (" << autorestart << ") t=" << cur_t
                                   << " nstp= " << get_n_steps() << " nfev=" << get_n_rhs_evals() << " ae=" << ((autonomous_exprs) ? 't' : 'f') << "\n";
                         if (status >= 0) {
-                            N_Vector ele, ew;
-                            ele = N_VNew_Serial(ny);
-                            ew = N_VNew_Serial(ny);
+                            SVector ele{ny
+#if SUNDIALS_VERSION_MAJOR >= 6
+                                  , *ctx
+#endif
+                            }, ew {
+ny
+#if SUNDIALS_VERSION_MAJOR >= 6
+                                  , *ctx
+#endif
+                                };
                             get_est_local_errors(ele);
                             get_err_weights(ew);
                             realtype mx = 0.0;
                             int mxi = -1;
                             for (unsigned i=0; i < ny; ++i){
-                                const realtype cur = NV_DATA_S(ele)[i]*NV_DATA_S(ew)[i];
+                                const realtype cur = ele[i]*ew[i];
                                 if (cur > mx){
                                     mxi = i;
                                     mx = cur;
                                 }
                             }
-                            N_VDestroy_Serial(ele);
-                            N_VDestroy_Serial(ew);
                             std::clog << "cvodes_cxx.hpp:" << __LINE__ << ":" << __FUNCTION__ << ":     max(ew[i]*ele[i]) = " << mx << ", i=" << mxi << "\n";
                         }
                     }
@@ -1505,7 +1837,7 @@ public:
                     if (constraints_.size()) {
                         for (int i=0; i<ny; ++i){
                             if (constraints_[i] == 1.0 and yqout[i + (iout-1)*((nderiv+1)*ny+nq)] < 0) {
-                                if (this->verbosity > 60) { std::clog << "clipping to y[" << i << "] to zero.\n"; }
+                                if (this->verbosity > 40) { std::clog << "clipping to y[" << i << "] to zero.\n"; }
                                 yqout[i + (iout-1)*((nderiv+1)*ny+nq)] = 0;
                             }
                         }
@@ -1526,6 +1858,9 @@ public:
                 }
             }
         }
+        if (this->verbosity > 0) {
+            std::fflush(stderr);
+        }
         return iout;
     }
 
